# Copyright(c) 2002 Association of Universities for Research in Astronomy, Inc.

procedure ggain (images)

# Multiply extensions by their gains to help remove chip-chip variations
# 
# Version   Feb 28, 2002 BM  v1.3 release

string images       {"",prompt="Input images"}
string gaindb       {"gmos$data/gmosamps.dat",prompt="Database with gain data"}
string logfile      {"",prompt="Logfile"}
string key_gain     {"GAIN",prompt="Header keyword for gain (e-/ADU)"}
string key_ron      {"RDNOISE",prompt="Header keyword for readout noise"}
real   gain         {2.2,prompt="Gain value to use if gain not found in db"}
real   ron          {3.5,prompt="Readout noise value to use if ron not found in db"}
bool  fl_update     {yes,prompt="Update headers?"}
bool  fl_mult       {no,prompt="Multiply by the gains?"}
string gainout      {"",prompt="Gains found"}
string ronout       {"",prompt="Read noise found"}
string sci_ext      {"SCI",prompt="Name of science extension"}
string var_ext      {"VAR",prompt="Name of variance extension"}
bool   verbose      {yes,prompt="Verbose?"}
int    status       {0,prompt="Exit status (0=good)"}
struct *scanfile    {"",prompt="Internal use only"}

begin

string l_images,l_gaindb,l_logfile
string l_key_gain, l_key_ron
string l_sci_ext,l_var_ext
string l_readmode, l_gainmode,dum
string img,inimages[200],filelist
real l_gain,l_rdnoise
int i,j,n,nbad,nsci
bool l_verbose,l_fl_mult,l_fl_update
struct amp,sdate

l_images=images
l_logfile=logfile
l_sci_ext=sci_ext ; l_var_ext=var_ext
l_gaindb=osfn(gaindb)
l_key_gain=key_gain ; l_key_ron=key_ron
l_verbose=verbose
l_fl_mult=fl_mult ; l_fl_update=fl_update
if (l_fl_mult) {
    l_fl_update=yes
}

cache("imgets","gimverify","keypar")
status=0

# Test the logfile:
if (l_logfile == "" || stridx(" ",l_logfile)>0) {
    l_logfile = gmos.logfile
    if (l_logfile == "" || stridx(" ",l_logfile)>0) {
        l_logfile = "gmos.log"
        printlog("WARNING - GGAIN: Both ggain.logfile and gmos.logfile fields are empty",l_logfile,yes)
        printlog("                    Using default file gmos.log",l_logfile,yes)
    }
}

date | scan(sdate)
printlog("--------------------------------------------------------------------------------",l_logfile,verbose=l_verbose)
printlog("GGAIN -- "//sdate,l_logfile,verbose=l_verbose)

# Test input images
if (l_images == "" || stridx(" ",l_images)>0) {
    printlog("ERROR - GGAIN: No input images given",l_logfile,yes)
    goto error
}

# Test gaindb
if (!access(l_gaindb)) {
    printlog("ERROR - GGAIN: Gain database not found",l_logfile,yes)
    goto error
}

# Test for @filelist
if (substr(l_images,1,1) == "@" && !access(substr(l_images,2,strlen(l_images))) ) {
   printlog("ERROR - GGAIN: Input list "//substr(l_images,2,strlen(l_images))//" does not exist",l_logfile,yes)
   goto error
}

filelist=mktemp("tmpfiles")

files(l_images, sort-, > filelist)
scanfile=filelist
n=0
nbad=0
while(fscan(scanfile, img) != EOF) {
    n=n+1
    gimverify(img)
    if (gimverify.status != 0) {
        nbad=nbad+1
    } else {
        inimages[n]=gimverify.outname//".fits"
    }
    keypar(img//"[0]","GGAIN",silent+)
    if (keypar.found==yes) {
        printlog("WARNING - GGAIN: "//img//" already gain corrected",l_logfile,l_verbose)
        nbad=nbad+1
    }
}
scanfile=""

if (nbad>0) {
    printlog("ERROR - GGAIN: "//nbad//" errors or images already gain corrected.",l_logfile,yes)
    goto error
}


for (i=1; i<=n; i+=1) {
    #print(inimages[i])
    imgets(inimages[i]//"[0]","NSCIEXT", >>& "dev$null")
    nsci=int(imgets.value)
    if (nsci==0) {
        printlog("ERROR - GGAIN: keyword NSCIEXT does not exist, please run gprepare.",l_logfile,yes)
        goto error
    }

    # Get the readmode and set the gains
    l_readmode="slow" ; l_gainmode="low"
    keypar(inimages[i]//"[0]","AMPINTEG",silent+)
    if(keypar.found==no) {
        printlog("WARNING - GGAIN: Cannot find CCD readmode for image "//input[ii],l_logfile,verbose+)
        printlog("                    Assuming slow read",l_logfile,verbose+)
    } else if (keypar.value=="1000") {
        l_readmode="fast"
    }
    
    imgets(inimages[i]//"["//l_sci_ext//",1]","GAIN", >>& "dev$null")
    if( real(imgets.value)>4. ) {
        l_gainmode="high"
    } if( real(imgets.value)<4. ) {
        l_gainmode="low"
    }

    for (j=1; j<=nsci; j+=1) {
        date | scan(sdate)
        imgets(inimages[i]//"["//l_sci_ext//","//j//"]","AMPNAME")
        l_gain=0.0 ; l_rdnoise=0.0
        match(l_readmode,l_gaindb,stop-) | \
        match(l_gainmode,"STDIN",stop-) | \
        match(imgets.value,"STDIN",stop-) | scan(dum,dum,l_gain,l_rdnoise,amp)
        printlog(inimages[i]//"["//l_sci_ext//","//j//"]: "//amp//" "//l_gain//" "//l_rdnoise,l_logfile,l_verbose)
        if (l_gain==0.0) {
            printlog("WARNING - GGAIN: gain and readnoise not found in "//l_gaindb//" for this mode",l_logfile,l_verbose)
            l_gain=gain ; l_rdnoise=ron
            printlog("Using values from parameter",l_logfile,l_verbose)
        }
        if (j==1) {
            gainout=str(l_gain)
            ronout=str(l_rdnoise)
        } else {
            gainout=gainout//" "//str(l_gain)
            ronout=ronout//" "//str(l_rdnoise)
        }
        # update header
        if (l_fl_update) {
            hedit(inimages[i]//"["//l_sci_ext//","//j//"]",l_key_gain,l_gain,update+,verify-,show-)
            hedit(inimages[i]//"["//l_sci_ext//","//j//"]",l_key_ron,l_rdnoise,update+,verify-,show-)
        }
        # multiply science extension by the gain
        if (l_fl_mult) {
            imarith(inimages[i]//"["//l_sci_ext//","//j//"]","*",l_gain,inimages[i]//"[SCI,"//j//",overwrite+]",verb-)
            hedit(inimages[i]//"["//l_sci_ext//","//j//"]",l_key_gain,1.0,update+,verify-,show-)
            gemhedit(inimages[i]//"["//l_sci_ext//","//j//"]","GAINMULT",l_gain,"Gain multiplication")
            gemhedit(inimages[i]//"[0]","GGAIN",sdate,"Date of GGAIN processing")
            hedit(inimages[i]//"[0]","GEM-TLM",sdate,update+,verify-,show-)
        } 
        # If a variance plane, multiply by gain
        if (imaccess(inimages[i]//"[VAR,"//j//"]")) {
            if (l_fl_update) {
            hedit(inimages[i]//"["//l_var_ext//","//j//"]",l_key_gain,l_gain,update+,verify-,show-)
            hedit(inimages[i]//"["//l_var_ext//","//j//"]",l_key_ron,l_rdnoise,update+,verify-,show-)
            #gemhedit(inimages[i]//"["//l_var_ext//","//j//"]","GAINMULT",l_gain,"Input gain ")
            }
            if (l_fl_mult) {
            imarith(inimages[i]//"["//l_var_ext//","//j//"]","*",l_gain,inimages[i]//"[SCI,"//j//",overwrite+]",verb-)
            hedit(inimages[i]//"["//l_var_ext//","//j//"]",l_key_gain,1.0,update+,verify-,show-)
            gemhedit(inimages[i]//"["//l_var_ext//","//j//"]","GAINMULT",l_gain,"Gain multiplication ")
            }
        }
    }
}

goto clean

error:
    status=1

clean:
    delete(filelist,verify-, >>& "dev$null")
    # close log file
    printlog(" ",l_logfile,l_verbose)
    printlog("GGAIN done",l_logfile,l_verbose)
    printlog("--------------------------------------------------------------------------------",l_logfile,l_verbose )

end
