# Copyright(c) 2002 Association of Universities for Research in Astronomy, Inc

procedure giflat(inflats,outflat)

# Derive normalize imaging flat for GMOS
#
# Version  Feb 28, 2002  MT,IJ,BM  v1.3 release
#

char   inflats      {"",prompt="Input flat field images"}
char   outflat      {"",prompt="Output flat field image"}
char   normsec      {"default",prompt="Image section to get the normalization."} 
bool   fl_scale     {yes,prompt="Scale the flat images before combining?"}
char   sctype       {"mean",prompt="Type of statistics to compute for scaling",
enum="mean|mode|midpt"}
char   statsec      {"default",prompt="Image section for relative intensity \
scaling"}
char   key_gain     {"GAIN",prompt="Header keyword for gain (e-/ADU)\n"}

char   sci_ext      {"SCI",prompt="Name of science extension"}
char   var_ext      {"VAR",prompt="Name of variance extension"}
char   dq_ext       {"DQ",prompt="Name of data quality extension"}
bool   fl_vardq     {no,prompt="Create variance and data quality frames?"}
int    sat          {65000, prompt="Saturation level in raw images (ADU)\n"}
bool   verbose      {yes,prompt="Verbose output?"}
char   logfile      {"",prompt="Name of logfile"}
int    status       {0,prompt="Exit status (0=good)"}

char   combine      {"average",prompt="Type of combine operation",
enum="average|median"}
char   reject       {"avsigclip",prompt="Type of rejection in flat average",
enum="none|minmax|avsigclip"}
real   lthreshold   {INDEF,prompt="Lower threshold when combining"}
real   hthreshold   {INDEF,prompt="Upper threshold when combining"}
int    nlow         {0,min=0,prompt="minmax: Number of low pixels to reject"}
int    nhigh        {1,min=0,prompt="minmax: Number of high pixels to reject"}
int    nkeep        {1,prompt="avsigclip: Minimum to keep (pos) or maximum \
to reject (neg)"}
bool   mclip        {yes,prompt="avsigclip: Use median in clipping algorithm?"}
real   lsigma       {3.,prompt="avsigclip: Lower sigma clipping factor"}
real   hsigma       {3.,prompt="avsigclip: Upper sigma clipping factor"}
real   sigscale     {0.1,prompt="avsigclip: Tolerance for clipping scaling \
corrections"}
real   grow         {0.,prompt="minmax or avsigclip: Radius (pixels) for \
neighbor rejection\n"}

char   gp_outpref   {"g",prompt="Gprepare prefix for output images"}
char   rawpath      {"",prompt="GPREPARE: Path for input raw images"}
char   key_ron      {"RDNOISE",prompt="Header keyword for readout noise"}
char   key_datasec  {"DATASEC",prompt="Header keyword for data section"}
char   gaindb       {"gmos$data/gmosamps.dat",prompt="Database with gain data"}
char   bpm          {"",prompt="Bad pixel mask\n"}

char   gi_outpref   {"r",prompt="Gireduce prefix for output images"}
char   bias         {"",prompt="Bias calibration image"}
bool   fl_over      {no,prompt="Subtract overscan level?"}
bool   fl_inter     {no,prompt="Interactive overscan fitting?"}
char   key_biassec  {"BIASSEC",prompt="Header keyword for overscan image \
section"}
bool   median       {no,prompt="Use median instead of average in column bias?"}
char   function     {"chebyshev",prompt="Overscan fitting \
function.",enum="legendre|chebyshev|spline1|spline3"}
int    order        {1,prompt="Order of overscan fitting function."}
real   low_reject   {3.,prompt="Low sigma rejection factor."}
real   high_reject  {3.,prompt="High sigma rejection factor."}
int    niterate     {2,prompt="Number of rejection iterations.\n"}

struct* scanfile    {"",prompt="Internal use only"} 

################################################################################
begin
################################################################################
char  l_inflats, l_outflat, l_bias, l_bpm, l_normsec
char  l_sci_ext, l_var_ext, l_dq_ext, l_combine, l_sctype, l_statsec, l_reject
char  l_key_biassec, l_key_datasec, l_gp_outpref, l_key_ron, l_key_gain
char  l_gi_outpref, l_function, l_logfile, l_gaindb,l_rawpath
bool  l_fl_vardq, l_fl_scale, l_mclip, l_fl_inter, l_median
bool  l_verbose, l_fl_over
real  l_lthreshold, l_hthreshold, l_lsigma, l_hsigma, l_sigscale, l_grow
real  l_low_reject, l_high_reject
int   l_nlow, l_nhigh, l_nkeep, l_sat, l_order, l_niterate

# Script variables
int   nbad, nfiles, ii, jj, nccd, sx1, sx2, sy1, sy2, nx1, nx2, ny1, ny2
real  l_gains[6], l_gainsl[6], l_gainfl[6], l_gainfh[6], norm, stddev, gain_norm 
real  gain1, gain2, gain3, gain4, gain5, gain6
real  meanval, firstval, newgain
char  s_empty, img, trimsec, datasec, ccdsum[6], fccd[6]
char  l_readmode, ll_readmode, l_gainmode, ll_gainmode
char  tmpfile, tmpinlist, varout[6]
char  tmpraw, tmpgprep, tmpgired
char  obsm, masktyp, scaletype, scheck
bool  fl_gprepare, badformat, ocreate
bool  ock,bck,fck,tck
struct  l_struct

char  gaindata

################################################################################
cache("imgets","gimverify","keypar","fparse")

################################################################################
# Set local variables
################################################################################
l_inflats = inflats; l_outflat = outflat; l_bias = bias; l_bpm = bpm
l_normsec = normsec; l_sci_ext=sci_ext; l_var_ext=var_ext; l_dq_ext=dq_ext
l_fl_vardq=fl_vardq; l_combine = combine; l_fl_scale = fl_scale
l_sctype=sctype; l_statsec=statsec; l_reject=reject;  l_lthreshold=lthreshold
l_hthreshold=hthreshold; l_nlow=nlow; l_nhigh=nhigh; l_nkeep=nkeep
l_mclip=mclip; l_lsigma=lsigma; l_hsigma=hsigma; l_sigscale=sigscale
l_grow=grow; l_key_biassec=key_biassec; l_key_datasec=key_datasec
l_gp_outpref=gp_outpref; l_key_ron=key_ron; l_key_gain=key_gain; l_sat = sat
l_gi_outpref=gi_outpref
l_fl_inter=fl_inter; l_median=median; l_function=function; l_order=order
l_low_reject=low_reject; l_high_reject=high_reject; l_niterate=niterate
l_verbose=verbose; l_logfile=logfile; l_gaindb=gaindb; l_rawpath=rawpath
l_fl_over=fl_over

###############################################################################
# Initialize variables
################################################################################
status = 0
nccd = 3
ock=no;bck=no;tck=no;fck=no

################################################################################
# Make temporary files
###############################################################################
tmpfile = mktemp("tmpfile")
tmpinlist = mktemp("tmpinlist")
tmpraw = mktemp("tmpraw")
tmpgprep = mktemp("tmpgprep")
tmpgired = mktemp("tmpgired") 

if(l_fl_vardq) {
     for(ii=1; ii<=nccd; ii+=1) {
          varout[ii]=mktemp("tmpsig")
     }
} else {
     for(ii=1; ii<=nccd; ii+=1) {
          varout[ii]=""
     }     
}               

###############################################################################
# Check string parameters
#############################################################################
################################################################################
# The logfile:
################################################################################
s_empty=""; print(l_logfile) | scan(s_empty); l_logfile=s_empty
if (l_logfile == "") {
    l_logfile = gmos.logfile
    if (l_logfile == "") {
        l_logfile = "gmos.log"
        printlog("WARNING - GIFLAT: both giflat.logfile and gmos.logfile are \
          empty.",l_logfile,verbose+)
        printlog("                Using default file gmos.log.",l_logfile,
          verbose+)
    }
}

# Now start logging: (there are 76 dashes)
date | scan(l_struct)
printlog("-----------------------------------------------------------------\
-----------",l_logfile,l_verbose)
printlog("GIFLAT -- "//l_struct,l_logfile,l_verbose)
printlog(" ",l_logfile,l_verbose)
printlog("Flat field images = "//l_inflats,l_logfile,l_verbose)
printlog("Output flat field image = "//l_outflat,l_logfile,l_verbose)
printlog(" ",l_logfile,l_verbose)

################################################################################
# Check that output flat field image does not exist.
gimverify(l_outflat)
if(gimverify.status!=1) {
     ocreate = no
     printlog("ERROR - GIFLAT: Output flat field image ("//l_outflat//") \
           already exists.",l_logfile,verbose+)
     goto crash
} 
l_outflat = gimverify.outname
ocreate = yes

################################################################################
# Check for empty strings:
################################################################################
# SCI and DQ extensions - 
s_empty=""; print(l_sci_ext) | scan(s_empty); l_sci_ext=s_empty
if(l_sci_ext=="") {
     printlog("ERROR - GIFLAT: Science extension name sci_ext is missing",
          l_logfile,verbose+)
     goto crash
}
s_empty=""; print(l_dq_ext) | scan(s_empty); l_dq_ext=s_empty
if(l_dq_ext=="") {
     printlog("ERROR - GIFLAT: Data quality extension name dq_ext is missing",
          l_logfile,verbose+)
     goto crash
}

################################################################################
# Check input and output files:
################################################################################
# Input files
if(substr(l_inflats,1,1)=="@") {
  img=substr(l_inflats,2,strlen(l_inflats))
  if(!access(img)) {
    printlog("ERROR - GIFLAT:  Input file "//img//" not found.",
     l_logfile,verbose+)
    goto crash
  }
}

files(l_inflats,sort-, > tmpfile)

nbad=0; nfiles=0; fccd[1]=""; fccd[2]=""; fccd[3]=""
scanfile = tmpfile
while(fscan(scanfile,img) != EOF) {

     # Check if image exists
     gimverify(img)
     if(gimverify.status==0) {
          img = gimverify.outname//".fits"
          keypar(img//"[0]","GIFLAT",silent+)
          if (keypar.found) {
               printlog("WARNING - GIFLAT: Image "//img//" has been \
                         run through giflat before.",l_logfile,verbose+)
               goto nextimage   
          }
          
          nfiles+=1
          
          # Check if image has not been gprepare'd
          keypar(img//"[0]","GPREPARE",silent+)
          if(!keypar.found) {
               print(img, >> tmpgprep)
               goto nextimage
          }
                    
          # Check if image has not been gireduce'd
          keypar(img//"[0]","GIREDUCE",silent+)
          if(!keypar.found) {
               gimverify(l_gi_outpref//img)
               if(gimverify.status==0) {
                    keypar(l_gi_outpref//img//"[0]","GIREDUCE",silent+)
                    if(keypar.found) print(gimverify.outname, >> tmpinlist)
               } else print(img, >> tmpgired) 
               goto nextimage               
          } 
          print(img, >> tmpinlist)
          goto nextimage          
     }
     
     # Check if image with prefixes exist - gprepare'd and gireduce'd
     gimverify(l_gi_outpref//l_gp_outpref//img)
     if(gimverify.status==0) {
          img = gimverify.outname//".fits"
          keypar(img//"[0]","GIREDUCE",silent+)
          if(!keypar.found) {
               printlog("ERROR - GIFLAT: prefixed image exists: "//img//" \
                    but has not been gireduce'd.",l_logfile,l_verbose)
               goto nextimage
          }                                                  
          print(img, >> tmpinlist)
          nfiles += 1
          printlog("WARNING - GIFLAT: using previously gireduce'd image: "//img,
               l_logfile,l_verbose)
          goto nextimage
     }
    

     
     # Check if gprepared image with prefix exists.
     gimverify(l_gp_outpref//img)
     if(gimverify.status==0) {
          img = gimverify.outname//".fits"
          keypar(img//"[0]","GPREPARE",silent+)
          if(!keypar.found) {
               printlog("ERROR - GIFLAT: prefixed image exists: "//img//" \
                    but has not been gprepare'd.",l_logfile,l_verbose)
               goto nextimage
          }                                                  
          print(img, >> tmpgired)
          nfiles += 1
          printlog("WARNING - GIFLAT: using previously gprepare'd: "//img,
               l_logfile,l_verbose)
          goto nextimage
     }
    
    
     # Check if image is "raw".
     gimverify(l_rawpath//img)
     if(gimverify.status==0) {
          keypar(l_rawpath//img//"[0]","GPREPARE",silent+)
          if (!keypar.found) {
               print(img, >> tmpgprep)
               nfiles +=1
               goto nextimage
          }               
               
          keypar(l_rawpath//img//"[0]","GIREDUCE",silent+)
          if (!keypar.found) {
               print(img, >> tmpgired)
               nfiles +=1
               goto nextimage
          }
          
          keypar(l_rawpath//img//"[0]","GIFLAT",silent+)
          if (keypar.found) {
               printlog("WARNING - GIFLAT: Image "//img//" has been \
                         run through giflat before.",l_logfile,verbose+)
               goto nextimage
          }
                                                          
     } else if(gimverify.status==1) {
          printlog("ERROR - GIFLAT: Input file "//img//" not found.",l_logfile,
               verbose+)
          nbad+=1
          goto nextimage
     } else if(gimverify.status>1) {
          printlog("ERROR - GIFLAT: Input file "//img//" not a MEF FITS image.",
               l_logfile,verbose+)
          nbad+=1
          goto nextimage
     } 
     nextimage:
}
     

# Check for empty file list
if(nfiles==0) {
  printlog("ERROR - GIFLAT:  No input images to process.",
     l_logfile,verbose+)
  goto crash
}

# Exit if problems found with input files
if (nbad > 0) {
   printlog("ERROR - GIFLAT: "//nbad//" image(s) either do not exist or \
     are not MEF files.",l_logfile, verbose+)
   goto crash
} # end if (nbad > 0)

if(access(tmpgprep)) {
     printlog("Running gprepare on input files:",l_logfile,l_verbose)
     if(l_verbose) type(tmpgprep)
     type(tmpgprep, >> l_logfile)

     # RDNOISE and GAIN keywords - 
     s_empty=""; print(l_key_ron) | scan(s_empty); l_key_ron=s_empty
     if (l_key_ron == "") {
          printlog ("WARNING - GIFLAT: Readout noise keyword parameter \
               key_ron is missing. Using default values.",l_logfile,verbose+)
     }
     s_empty=""; print(l_key_gain) | scan(s_empty); l_key_gain=s_empty
     if (l_key_gain == "") {
          printlog ("WARNING - GIFLAT: Gain keyword parameter key_gain is \
               missing. Using default values.",l_logfile,verbose+)
     } 

     # Input BPM file -
     if(l_fl_vardq) {
          s_empty=""; print(l_bpm) | scan(s_empty); l_bpm=s_empty
          if(l_bpm=="") {
               printlog("WARNING - GIFLAT: Bad pixel mask is an empty string",
                    l_logfile,verbose+)
          } else {
               gimverify(l_bpm); l_bpm=gimverify.outname

               if(gimverify.status==1) {
                    printlog("ERROR - GIFLAT: Bad pixel mask "//l_bpm//" not \
                         found.",l_logfile,verbose+)
                    goto crash
               } else if(gimverify.status>1) {
                    printlog("ERROR - GIFLAT: Bad pixel mask "//l_bpm//" not \
                         a MEF FITS image.",l_logfile,verbose+)
               } else {
                    imgets(l_bpm//"[0]","GBPM", >& "dev$null")
                    if (imgets.value == "0") {
                         printlog("WARNING - GIFLAT: Bad pixel \
                              mask "//l_bpm//" was not created with GMOS GBPM \
                              task.",l_logfile,verbose+)
                    }
                                                                       
               }
          }
     }

     gprepare("@"//tmpgprep,rawpath=l_rawpath,outimages="",
          bpm=l_bpm,fl_addm=no,logfile=l_logfile,outpref=l_gp_outpref,
          fl_vardq=l_fl_vardq,sci_ext=l_sci_ext,var_ext=l_var_ext,
          dq_ext=l_dq_ext,key_ron=l_key_ron,key_gain=l_key_gain,
          gaindb=l_gaindb,key_dsec=l_key_datasec,sat=l_sat,verbose=l_verbose)
                    
     if( gprepare.status != 0) nbad+=1
     else sections(l_gp_outpref//"@"//tmpgprep,option="fullname", >> tmpgired)
}         
  

if(access(tmpgired)) {

     # Input bias image -
     s_empty=""; print(l_bias) | scan(s_empty); l_bias=s_empty
     gimverify(l_bias); l_bias=gimverify.outname
     if(gimverify.status!=0) {
          printlog("ERROR - GIFLAT: Bias image ("//l_bias//") does not \
               exist.",l_logfile, verbose+)
          goto crash 
     }
     printlog("Running gireduce on input files:",l_logfile,l_verbose)
     if(l_verbose) type(tmpgired)
     type(tmpgired, >> l_logfile)
     
     gireduce("@"//tmpgired,outpref=l_gi_outpref,outimages="",sci_ext=l_sci_ext,
          var_ext=l_var_ext,dq_ext=l_dq_ext,fl_vardq=l_fl_vardq,
          fl_over=l_fl_over,key_biassec=l_key_biassec,median=l_median,
          fl_inter=l_fl_inter,function=l_function,order=l_order,
          low_reject=l_low_reject,high_reject=l_high_reject,
          niterate=l_niterate,fl_trim+,key_datasec=l_key_datasec,fl_bias+,
          bias=l_bias,fl_dark-,fl_flat-,logfile=l_logfile,
          verbose=l_verbose,gaindb=l_gaindb,rawpath=l_rawpath)
     if(gireduce.status!=0) {
          printlog("ERROR - GIFLAT: gireduce returned with error.",l_logfile,
               yes)
          nbad+=1
     } else sections(l_gi_outpref//"@"//tmpgired,option="fullname", >> tmpinlist)

}
delete(tmpgprep,ver-, >& "dev$null")
delete(tmpgired,ver-, >& "dev$null")

nbad=0
head(tmpinlist,nlines=1) | scan(img)
for(ii=1; ii<=3; ii+=1) {
     imgets(img//"["//l_sci_ext//","//ii//"]","CCDSUM", >& "dev$null")
     fccd[ii]=imgets.value; ccdsum[ii]=fccd[ii]
}
count(tmpinlist) | scan(nfiles)
delete(tmpfile,ver-)
tail(tmpinlist,nlines=nfiles-1, > tmpfile)
scanfile = tmpfile
while(fscan(scanfile,img) != EOF) {

     for(ii=1; ii<=3; ii+=1) {
          imgets(img//"["//l_sci_ext//","//ii//"]","CCDSUM", >& "dev$null")
          ccdsum[ii]=imgets.value
     }

     if(ccdsum[1]!=fccd[1] || ccdsum[2]!=fccd[2] || ccdsum[3]!=fccd[3]) {
          printlog("ERROR - GIFLAT: input images have different \
               binning.",l_logfile,yes)
          nbad+=1
     }
                                                  
} # end while loop
# Exit if problems found with CCDSUM 
if (nbad > 0) goto crash

printlog("Using input files:",l_logfile,l_verbose)
if(l_verbose) type(tmpinlist)
type(tmpinlist, >> l_logfile)
delete(tmpfile,ver-)
################################################################################
# Update the default for binning:
# Check l_statsec
badformat=no
s_empty=""; print(l_statsec) | scan(s_empty); l_statsec=s_empty
print(l_statsec) | ucase | scan(l_statsec)
ii = strlen(l_sci_ext)
if(l_statsec=="DEFAULT") {
     l_statsec = "["//l_sci_ext//",2][100:1900,100:4500]"
     if(ccdsum[1]=="2 2" && ccdsum[2]=="2 2" && ccdsum[3]=="2 2")
          l_statsec = "["//l_sci_ext//",2][50:800,50:2250]"
} else if (substr(l_statsec,2,4) != l_sci_ext) {
     printlog("ERROR - GIFLAT: illegal parameter STATSEC ("//l_statsec//") \
          format is e.g [SCI,2][100:1900,100:4500]",l_logfile,yes)
     badformat=yes
}
scheck = l_statsec
ii=stridx("]",scheck);jj=stridx(":",scheck);sx1=int(substr(scheck,ii+2,jj-1))
scheck = substr(scheck,jj+1,strlen(scheck))
jj=stridx(",",scheck); sx2=int(substr(scheck,1,jj-1))
ii=stridx(":",scheck); sy1=int(substr(scheck,jj+1,ii-1))
sy2=int(substr(scheck,ii+1,strlen(scheck)-1))
if(sx1>sx2 || sy1>sy2 || sx1<0 || sx2<0) {
     printlog("ERROR - GIFLAT: Illegal parameter STATSEC: "//l_statsec,
          l_logfile,yes)
     badformat=yes
}     
     
# Check l_normsec
s_empty=""; print(l_normsec) | scan(s_empty); l_normsec=s_empty
print(l_normsec) | ucase | scan(l_normsec)
if(l_normsec=="DEFAULT") {
     l_normsec = "["//l_sci_ext//",2][100:1900,100:4500]"
     if(ccdsum[1]=="2 2" && ccdsum[2]=="2 2" && ccdsum[3]=="2 2")
          l_normsec = "["//l_sci_ext//",2][50:800,50:2250]"     
} else if (substr(l_normsec,2,4) != l_sci_ext) {
     printlog("ERROR - GIFLAT: illegal parameter NORMSEC ("//l_normsec//") \
          format is e.g [SCI,2][100:1900,100:4500]",l_logfile,yes)
     badformat=yes
}     
scheck = l_normsec
ii=stridx("]",scheck);jj=stridx(":",scheck);nx1=int(substr(scheck,ii+2,jj-1))
scheck = substr(scheck,jj+1,strlen(scheck))
jj=stridx(",",scheck); nx2=int(substr(scheck,1,jj-1))
ii=stridx(":",scheck); ny1=int(substr(scheck,jj+1,ii-1))
ny2=int(substr(scheck,ii+1,strlen(scheck)-1))
if(nx1>nx2 || ny1>ny2 || nx1<0 || nx2<0) {
     printlog("ERROR - GIFLAT: Illegal parameter NORMSEC: "//l_normsec,
          l_logfile,yes)
     badformat=yes
}     
 
if(badformat) goto crash
printlog("Using STATSEC: "//l_statsec,l_logfile,l_verbose)
printlog("Using NORMSEC: "//l_normsec,l_logfile,l_verbose)

################################################################################
# Create an empty MEF flat field -
imcopy(img//"[0]",l_outflat//".fits",verbose-)
################################################################################
# The input flats are in tmpinlist
# The output flat is named: l_outflat
################################################################################
# Check that for flat field images OBSMODE=IMAGE and MASKTYPE=0
################################################################################
rename(tmpinlist,tmpfile,field="all")
scanfile = tmpfile
l_readmode="slow"  # default readmode
l_gainmode="low"   # default gainmode
ii=1
while(fscan(scanfile,img) != EOF ) {
     imgets(img//"[0]","OBSMODE", >& "dev$null"); obsm = imgets.value
     imgets(img//"[0]","MASKTYP", >& "dev$null"); masktyp = imgets.value
     if (obsm == "IMAGE" && masktyp == "0") 
          print(img, >> tmpinlist)
     else 
          printlog("WARNING - GIFLAT: Input flat image "//img//" is not an \
               imaging flat.",l_logfile,verbose+)

     imgets(img//"[0]","AMPINTEG", >>& "dev$null")  
        if( real(imgets.value)==5000. )
             ll_readmode="slow"
        if( real(imgets.value)==1000. )
             ll_readmode="fast"
        if(ii==1) {
            l_readmode=ll_readmode
        } else if (l_readmode!=ll_readmode) {
          printlog("ERROR - GIFLAT: Cannot mix slow and fast readout images",l_logfile,yes)
          goto crash
        }
     imgets(img//"[SCI,1]","GAIN", >>& "dev$null")  
        if( real(imgets.value)>4. )
             ll_gainmode="high"
        if( real(imgets.value)<4. )
             ll_gainmode="low"
        if(ii==1) {
            l_gainmode=ll_gainmode
        } else if (l_gainmode!=ll_gainmode) {
          printlog("ERROR - GIFLAT: Cannot mix high and low gain images",l_logfile,yes)
          goto crash
        }
        ii+=1

}
delete(tmpfile, ver-)     
if(!access(tmpinlist)) {
     printlog("ERROR - GIFLAT: No flat field images available.",l_logfile,
          verbose+)
     goto crash
}
              
################################################################################
# Combine reduced images - they are in tmpinlist 
################################################################################
scaletype="none"
count(tmpinlist) | scan(nfiles)
if(l_fl_scale && nfiles>1) {
     scaletype="!RELINT"
     head(tmpinlist,nline=1) | scan(img)

     # Check the image size: statsec should be within the image
     imgets(img//"[SCI,2]","i_naxis1")
     if(int(imgets.value)<sx2) {
          printlog("ERROR - GIFLAT: Statistics section is outside image",
               l_logfile,yes)
          goto crash
     }
     imgets(img//"[SCI,2]","i_naxis2")
     if(int(imgets.value)<sy2) {
          printlog("ERROR - GIFLAT: Statistics section is outside image",
               l_logfile,yes)
          goto crash
     }               
     imstatistics(img//l_statsec,field="mean",lower=INDEF,upper=INDEF,
          binwidth=0.1,format-) | scan(firstval)
     for(ii=1; ii<=nccd; ii+=1) {                 
          hedit(img//"["//l_sci_ext//","//ii//"]","RELINT",1.,update+,add+,
               delete-,verify-,show-)
     }
     count(tmpinlist) | scan(ii)
     tail(tmpinlist,nlines=ii-1, > tmpfile)
     scanfile=tmpfile
     meanval=1.0
     while(fscan(scanfile,img) != EOF) {

          imstatistics(img//l_statsec,field=l_sctype,lower=INDEF,upper=INDEF,
               binwidth=0.1,format-) | scan(meanval)
          meanval = firstval/meanval 
          for(ii=1; ii<=nccd; ii+=1) {                 
               hedit(img//"["//l_sci_ext//","//ii//"]","RELINT",meanval,
                    update+,add+,delete-,verify-,show-)
          }
     }
     scanfile=""
     delete(tmpfile,ver-)     
}
  
# Run imcombine through all nccd science extensions
for(ii=1; ii<=nccd; ii+=1) {
     l_gains[ii]=0.0
     # Place all [sci,ii] extensions in file
     delete(tmpfile,ver-, >& "dev$null"); scanfile=tmpinlist
     while(fscan(scanfile,img) != EOF) {
         print(img//"["//l_sci_ext//","//ii//"]", >> tmpfile)
         if (l_gains[ii]==0.0) {
             imgets(img//"["//l_sci_ext//","//ii//"]",l_key_gain, >>& "dev$null")
             l_gains[ii]=real(imgets.value)
         }
     }
     imcombine("@"//tmpfile,l_outflat//"["//l_sci_ext//","//ii//",append]",
          rejmask="",plfile="",sigma=varout[ii],logfile=l_logfile,
          combine=l_combine,reject=l_reject,project-,outtype="real",
          offsets="none",masktype="none",blank=0.,scale=scaletype,zero="none",
          weight="none",statsec="",lthres=l_lthreshold,hthres=l_hthreshold,
          nlow=l_nlow,nhigh=l_nhigh,sigscale=l_sigscale,
          nkeep=l_nkeep,lsigma=l_lsigma,hsigma=l_hsigma,grow=l_grow)
     hedit("@"//tmpfile,"RELINT","",update+,add-,delete+,verify-,show-)
     delete(tmpfile,ver-) 
     # Change intensities to electrons - will do the var into electrons
     # conversion later.
     imarith(l_outflat//"["//l_sci_ext//","//ii//"]","*",l_gains[ii],
          l_outflat//"["//l_sci_ext//","//ii//",overwrite]",title="",hparams="",pixtype="",
          calctype="",verbose-,noact-)
     imgets(l_outflat//"["//l_sci_ext//","//ii//"]","NCOMBINE")
     hedit(l_outflat//"["//l_sci_ext//","//ii//"]","GAIN",real(imgets.value),
          update+,verify-,show-)
     gemhedit(l_outflat//"["//l_sci_ext//","//ii//"]","GAINORIG",l_gains[ii],
          "Orig. Gain [e-/ADU]")

}

# Create the data quality plane
if(fl_vardq) {
     head(tmpinlist,nlines=1) | scan(img)
     for(ii=1; ii<=nccd; ii+=1) {
          imcopy(img,l_outflat//"["//l_dq_ext//","//ii//",append]")
          scanfile=tmpinlist
          while(fscan(scanfile,img) != EOF) {
               # Quick and dirty - needs to use other task but this
               # will do if good pixels are 0 and bad pixels are
               # flagged as 1. 
               imarith(img,"max",l_outflat//"["//l_dq_ext//","//ii//"]",
                    l_outflat//"["//l_dq_ext//","//ii//",overwrite]",
                    title="",divzero=0.,hparams="",pixtype="",calctype="",
                    verbose-,noact-)
          }
     }
     
}
   
# Get the normalization from the center CCD
keypar(l_outflat//l_normsec,"GAINORIG",silent+)
newgain = real(keypar.value) 
keypar(l_outflat//l_normsec,"OVERSCAN",silent+)
if(keypar.found) l_sat = (l_sat - int(keypar.value))*newgain
else l_sat = l_sat*newgain
imstatistics(l_outflat//l_normsec,field="mode,stddev",
     lower=0.,upper=l_sat,format-) | scan(norm,stddev)
imstatistics(l_outflat//l_normsec,field="midpt",
     lower=(norm-3*stddev),upper=(norm+3*stddev),format-) | scan(norm)
printlog("GIFLAT: Normalization factor "//norm,l_logfile,l_verbose)

# Normalize the flats and change var into electrons with the norm correction
for(ii=1; ii<=nccd; ii+=1) {
     imarith(l_outflat//"["//l_sci_ext//","//ii//"]","/",norm,
          l_outflat//"["//l_sci_ext//","//ii//",overwrite]",title="",hparams="",pixtype="",
          calctype="",verbose-,noact-)          
     # Divide sigma by gain so that they are in electrons - note varout name is
     # really the sigma.     
     gain_norm = l_gains[ii]*norm 

     imarith(varout[ii],"/",gain_norm,varout[ii],title="",divzero=0.,
          hparams="",pixtype="",calctype="",verbose-,noact-)

     if(l_fl_vardq) {     
          # Square the sigma file to get the variance  
          imarith(varout[ii],"*",varout[ii],
               l_outflat//"["//l_var_ext//","//ii//",append]",
               title="",divzero=0.,hparams="",pixtype="",calctype="",verbose-,
               noact-)
     }               
}           

# Now update the PHU     
date | scan(l_struct)
gemhedit(l_outflat//"[0]","GIFLAT",l_struct,"Time stamp for GIFLAT")
gemhedit(l_outflat//"[0]","GIFLNORM",norm,"Normalization factor for GIFLAT")
gemhedit(l_outflat//"[0]","GEM-TLM",l_struct,"Last modification with GEMINI")

goto clean

################################################################################
# Exit with error
################################################################################
crash:
    status=1
    if(ocreate) imdelete(l_outflat,verify-, >& "dev$null")
    goto clean

################################################################################
# Clean up and exit
################################################################################
clean:
     delete(tmpfile//","//tmpinlist,ver-, >& "dev$null")
     delete(tmpgprep,ver-, >& "dev$null")
     delete(tmpgired,ver-, >& "dev$null")

     if(l_fl_vardq) {
          for(ii=1; ii<=6; ii+=1) {
               delete(varout[ii],ver-, >& "dev$null")
          }
     }               
    
    if (status == 0) {
       printlog("   ",l_logfile,l_verbose)
       printlog("GIFLAT exit status:  good.",l_logfile,l_verbose)
       printlog("----------------------------------------------------------\
       ------------------",l_logfile,l_verbose)
    } else {
       printlog("   ",l_logfile,l_verbose)
       printlog("GIFLAT exit status:  error.",l_logfile,l_verbose)
       printlog("-----------------------------------------------------------\
       -----------------",l_logfile,l_verbose)
    }
  
end
