# Copyright(c) 2002 Association of Universities for Research in Astronomy, Inc.

procedure gmosaic(inimages)

# Mosaic the 3 GMOS ccds
# geotran gdata$g24173[1][33:2080,1:4608] chip1 "" "" xshift=-2.50 yshift=-1.58 \
# xrotation=-0.0027827 yrotation=-0.0027827
#
# imcopy gdata$g24173[2][33:2080,1:4608] chip2
# geotran gdata$g24173[3][1:2048,1:4608] chip3 "" "" xshift=3.3723 yshift=-1.89 \
# xrotation=-0.056477 yrotation=-0.056477
# followed by imtile

# Version Feb 28, 2002  IJ,BM  v1.3 release

char inimages    {"",prompt="GMOS image to mosaic"}
char outimages   {"",prompt="Output images or list"}
char outpref     {"m",prompt="Prefix for output images"}
bool fl_paste    {no,prompt="Paste images instead of mosaic"}
int  gap         {37,prompt="Gap between the CCDs in unbinned pixels"}
char sci_ext     {"SCI",prompt="Extension(s) to mosaic, use '' for raw data"}
char key_detsec  {"DETSEC",prompt="Header keyword for detector section"}
char key_datsec  {"DATASEC",prompt="Header keyword for data section"}
char key_ccdsum  {"CCDSUM",prompt="Header keyword for CCD binning"}
char logfile     {"",prompt="Logfile"}
char bpmfile     {"gmos$data/chipgaps.dat",prompt="Info on location of chip gaps"}
bool fl_fixpix   {no,prompt="Interpolate across chip gaps?"}
bool verbose     {yes,prompt="Verbose"}
int  status      {0,prompt="Exit status (0=good)"}
struct* scanfile {"",prompt="Internal use only"}

begin

char l_inimages, l_outimages, l_outpref, l_logfile
char l_sci_ext, l_key_detsec, l_key_datsec, l_key_ccdsum, l_bpmfile
bool l_verbose, l_fl_paste,l_fl_fixpix
int l_gap, l_status

char tmpin, l_dsec, tmpout, l_paste, tmpbpm
char tmpim[3], l_imext
char l_pixtype, sciext, mdf
int n_ext, x1, x2, y1, y2, n_i, n_j, n_k
int nimages,nsci
int Xoff, Yoff, Xbin, Ybin
struct l_struct
char filelist, l_temp, l_inim[200], l_outim[200], l_datsec[6]
int maxfiles, nbad
int l_locgap  # local use of gap size
bool l_fl_bin

real lxshift[3], lyshift[3], lrotation[3], lxbshift[3]

# Transformation - unbinned pixels
#lxshift[1]=-2.50 ; lyshift[1]=-1.58 ; lrotation[1]=-0.0027827
# best for IFU
#lxshift[1]=-2.50 ; lyshift[1]=-1.62 ; lrotation[1]=-0.004 
# best overall
lxshift[1]=-2.50 ; lyshift[1]=-1.58 ; lrotation[1]=-0.004 
lxshift[2]=0. ; lyshift[2]=0. ; lrotation[2]=0. 
#lxshift[3]=3.3723 ; lyshift[3]=-1.89 ; lrotation[3]=-0.056477
# best for IFU (lower halves of chips)
#lxshift[3]=3.3723 ; lyshift[3]=-1.99 ; lrotation[3]=-0.054
#lxshift[3]=3.3723 ; lyshift[3]=-2.05 ; lrotation[3]=-0.056 
#lxshift[3]=3.3723 ; lyshift[3]=-1.89 ; lrotation[3]=-0.048
#best overall so far but need more than shift and rotation to match
lxshift[3]=3.3723 ; lyshift[3]=-1.86 ; lrotation[3]=-0.046

# Shift in X if binned pixels - still unbinned pixels in the shift
lxbshift[1]=-1.5 ; lxbshift[2]=0 ; lxbshift[3]=2.3723

l_inimages=inimages ; l_sci_ext=sci_ext ; l_key_detsec=key_detsec
l_outimages=outimages ; l_outpref=outpref
l_key_datsec=key_datsec ; l_key_ccdsum=key_ccdsum 
l_verbose=verbose ; l_gap=gap ; l_fl_paste=fl_paste
l_logfile=logfile 
l_fl_fixpix=fl_fixpix ; l_bpmfile=bpmfile
maxfiles=200
if (l_sci_ext != "") {
  l_sci_ext = l_sci_ext//","
}

tmpin="uparm$"//mktemp("tmpin")
tmpout=mktemp("tmpout")
tmpbpm=mktemp("tmpbpm")
tmpim[1]=mktemp("tmpim1")
tmpim[2]=mktemp("tmpim2")
tmpim[3]=mktemp("tmpim3")
filelist=mktemp("tmpfilelist")
mdf=mktemp("tmpmdf")
mdf=mdf//".fits"

# cheap and fast - for now
unlearn geotran
cache("imgets","gimverify","tinfo","hedit")
# Test the logfile:
if (l_logfile == "" || stridx(" ",l_logfile)>0) {
  l_logfile = gmos.logfile
  if (l_logfile == "" || stridx(" ",l_logfile)>0) {
     l_logfile = "gmos.log"
     printlog("WARNING - GMOSAIC: Both gmosaic.logfile and gmos.logfile fields are empty",l_logfile,yes)
     printlog("                   Using default file gmos.log",l_logfile,yes)
  }
}

date | scan(l_struct)
printlog("----------------------------------------------------------------------------", logfile=l_logfile, verbose=l_verbose)
printlog("GMOSAIC -- "//l_struct,l_logfile,l_verbose)
printlog("",l_logfile,l_verbose)
printlog("Input list = "//l_inimages,l_logfile,l_verbose)
printlog("Output list = "//l_outimages,l_logfile,l_verbose)
printlog("Output prefix = "//l_outpref,l_logfile,l_verbose)
printlog("",l_logfile,l_verbose)
printlog("Current version supports 3 amps only",
  l_logfile,l_verbose)

if (l_inimages == "" || stridx(" ",l_inimages)>0) {
  printlog("ERROR - GMOSAIC: Input file not specified",l_logfile,yes)
  goto crash
}

# parse wildcard and comma-separated lists
if(substr(l_inimages,1,1)=="@") {
  if(!access(substr(l_inimages,2,strlen(l_inimages))) ) {
   printlog("ERROR - GMOSAIC: Input list "//\
    substr(l_inimages,2,strlen(l_inimages))//" not found",l_logfile,yes)
   goto crash
  }
  scanfile=substr(l_inimages,2,strlen(l_inimages))
} else {
  files(l_inimages,sort-, > filelist)
  scanfile=filelist
}

nbad=0 ; nimages=0 
while(fscan(scanfile,l_temp) != EOF) {
  gimverify(l_temp)
  if(gimverify.status==1) {
    printlog("ERROR - GMOSAIC: Cannot access image "//l_temp,l_logfile,yes)
    nbad+=1
  } else if(gimverify.status>1) {
    printlog("ERROR - GMOSAIC: Image "//l_temp//" is not a MEF file",l_logfile,yes)
    nbad+=1
  } else {
    nimages+=1
    if(nimages>maxfiles) {
      printlog("ERROR - GMOSAIC: Maxfiles "//maxfiles//" exceeded",l_logfile,yes)
      goto crash
    }
    l_inim[nimages]=gimverify.outname
  }
}
scanfile=""
delete(filelist,ver-, >>& "dev$null")
if(nbad>0) {
  goto crash
}
if(l_outimages=="") {
# prefix
  nbad=0
  for(n_i=1;n_i<=nimages;n_i+=1) {
    l_outim[n_i]=l_outpref//l_inim[n_i]
    gimverify(l_outim[n_i])
    if(gimverify.status!=1) {
       printlog("ERROR - GMOSAIC: Output image "//l_outim[n_i]//" exists",l_logfile,yes)
       nbad+=1
    }
  }
  if(nbad>0) {
    goto crash
  }
# outimages
} else {
  if(substr(l_outimages,1,1)=="@") {
    scanfile=substr(l_outimages,2,strlen(l_outimages))
  } else {
    files(l_outimages,sort-, > filelist)
    scanfile=filelist
  }
  n_i=0
  while(fscan(scanfile,l_temp) != EOF) {
     n_i+=1
     gimverify(l_temp) ; l_outim[n_i]=gimverify.outname
     if(gimverify.status!=1) {
       printlog("ERROR - GMOSAIC: Output image "//l_outim[n_i]//" exists",l_logfile,yes)
       nbad+=1
     }
  }
  delete(filelist,ver-, >>& "dev$null")
  if(nbad>0) {
    goto crash
  }
  if(n_i!=nimages) {
     printlog("ERROR - GMOSAIC: Different number of input and output images",
     l_logfile,yes)
     goto crash
  } 
  
} # end of checking output images

# check bpm
if (l_fl_fixpix && !access(l_bpmfile)) {
   printlog("ERROR - GMOSAIC: bpmfile does not exist",l_logfile,yes)
   goto crash
}

if(l_fl_paste)
  printlog("Pasting images, not using full transformations",l_logfile,l_verbose)
else
  printlog("Mosaicing images, using full transformations",l_logfile,l_verbose)

# -----  The mosaicing -----

for(n_i=1;n_i<=nimages;n_i+=1) {
  l_fl_bin=no ; l_locgap=l_gap
  sciext="["//l_sci_ext//"1]"
# Check existence of extension
  if (!imaccess(l_inim[n_i]//sciext)) {
    printlog("ERROR - GMOSAIC: "//l_inim[n_i]//sciext//" not found",l_logfile,yes)
    goto crash
  }
 
# If there is a MDF, save it for later insertion in the final image
  tinfo(l_inim[n_i]//".fits[MDF]",ttout-, >& "dev$null")
  if (tinfo.tbltype=="fits") {
    printlog("Found a MDF, copying for later use.",l_logfile,l_verbose)
    tcopy(l_inim[n_i]//".fits[MDF]",mdf,verbose=no, >& "dev$null")
  }

# Get binning
  imgets(l_inim[n_i]//sciext,l_key_ccdsum)
  print(imgets.value) | \
    fields("STDIN","1,2",lines="1",quit_if_miss-,print_file-) | \
    scan(Xbin,Ybin)
  if(Xbin>1) {
    l_fl_bin=yes
    l_locgap=36.
  }

printlog("",l_logfile,l_verbose)
printlog("Input image: "//l_inim[n_i]//"  Output image: "//l_outim[n_i],
  l_logfile,l_verbose)

# Get number of science extensions
# imgets(l_inim[n_i],"NSCIEXT")
# nsci=int(imgets.value)
  nsci=3
# For each extension  - assume 3 for now 
  for(n_k=1;n_k<=nsci;n_k+=1) {
  sciext="["//l_sci_ext//str(n_k)//"]"
  imgets(l_inim[n_i]//sciext,l_key_datsec)
  l_datsec[n_k]=imgets.value

# Correct the rotation for the binning
if (Xbin>1 || Ybin>1) {
  lrotation[n_k] = lrotation[n_k] * Ybin/Xbin
}

# Setup parameters for imtile
  if(n_k==1) {
    l_locgap=nint(l_locgap/Xbin)
    if (l_fl_paste) {
      l_paste=l_inim[n_i]//sciext//l_datsec[n_k]
    }
    imgets(l_inim[n_i]//sciext,"i_pixtype")
    if (imgets.value=="11") {
      l_pixtype="l"
    } else {
      l_pixtype="r"
    }
  } else if(l_fl_paste) {
    l_paste=l_paste//","//l_inim[n_i]//sciext//l_datsec[n_k]
  }
  if(!l_fl_paste) {
# geotran
    if(n_k!=2) {
    if(l_fl_bin) {
      geotran(l_inim[n_i]//sciext//l_datsec[n_k],tmpim[n_k],"","",
      xshift=(lxbshift[n_k]/Xbin),yshift=(lyshift[n_k]/Ybin),
      xrotation=lrotation[n_k],yrotation=lrotation[n_k],
      xmag=1,ymag=1,xmin=INDEF,xmax=INDEF,ymin=INDEF,ymax=INDEF,
      ncols=INDEF,nlines=INDEF,verbose=l_verbose,fluxconserve=yes,
      nxblock=2048,nyblock=2048,interpolant="linear")
    } else {
      geotran(l_inim[n_i]//sciext//l_datsec[n_k],tmpim[n_k],"","",
      xshift=(lxshift[n_k]/Xbin),yshift=(lyshift[n_k]/Ybin),
      xrotation=lrotation[n_k],yrotation=lrotation[n_k],
      xmag=1,ymag=1,xmin=INDEF,xmax=INDEF,ymin=INDEF,ymax=INDEF,
      ncols=INDEF,nlines=INDEF,verbose=l_verbose,fluxconserve=yes,
      nxblock=2048,nyblock=2048,interpolant="linear")
    }
    if (n_k==1) {
        l_paste=tmpim[n_k]
    } else {
        l_paste=l_paste//","//tmpim[n_k]
    }
   }
   if (n_k==2) {
     l_paste=l_paste//","//l_inim[n_i]//sciext//l_datsec[n_k]
   }
  }  # end of !fl_paste
 } # end of running through the extensions
#print(l_paste)

# Get a PHU
  fxcopy(l_inim[n_i]//".fits",l_outim[n_i]//".fits",groups="0",new_file=yes,verbose=no, >>& "dev$null")
# Tile - append extension
  printlog("",l_logfile,l_verbose)
  imtile(l_paste,l_outim[n_i]//"[1,append]",3,1,ncoverlap=-l_locgap,trim_section="",
  missing_inpu="",start_tile="ll",row_order=yes,raster_order=no,
  median_secti="",subtract=no,ncols=INDEF,nlines=INDEF,nloverlap=-1,
  opixtype=l_pixtype,ovalue=0,verbose=l_verbose)
  hedit(l_outim[n_i]//"[1]","EXTNAME","SCI",add+,upd+,ver-,show-)
  hedit(l_outim[n_i]//"[1]","EXTVER","1",add+,upd+,ver-,show-)
  hedit(l_outim[n_i]//"[0]","NSCIEXT","1",add+,upd+,ver-,show-)
  hedit(l_outim[n_i]//"[0]","NEXTEND","1",add+,upd+,ver-,show-)
  flpr

# Fixpix if desired
  if (l_fl_fixpix) {
     printlog("",l_logfile,l_verbose)
     printlog("Interpolating across chip gaps using "//l_bpmfile,
      l_logfile,l_verbose)
     if(Xbin==1 && Ybin==1) {
        badpiximage(l_bpmfile,l_outim[n_i]//"[SCI,1]",tmpbpm//".pl",
         goodvalue=0, badvalue=1)
     } else {
        copy(l_bpmfile,tmpbpm,verbose-)
        if(Xbin>1) {
          tcalc(tmpbpm,"c1","int((c1-1.)/"//str(Xbin)//"+0.5)")
          tcalc(tmpbpm,"c2","int((c2-1.)/"//str(Xbin)//"+0.5)")
        }
        tcalc(tmpbpm,"c4","int(c4/"//str(Ybin)//")")
        badpiximage(tmpbpm,l_outim[n_i]//"[SCI,1]",tmpbpm//".pl",
         goodvalue=0, badvalue=1)
     }
     fixpix(l_outim[n_i]//"[SCI,1]",tmpbpm//".pl",linterp="",cinterp="",
     ninterp="1",creplace="")
     delete(tmpbpm//".pl,"//tmpbpm,verify-, >>& "dev$null")
  }

# add the MDF
  if (access(mdf)) {
    fxinsert(mdf//"[1]",l_outim[n_i]//".fits[1]","",verbose-, >& "dev$null")
    hedit(l_outim[n_i]//"[0]","NEXTEND",2,update+,verify-,show-)
    delete(mdf,verify-, >& "dev$null")
  }

# Get the WCS from ext [2] +shift - no other SCI extension info should be needed
#CTYPE1  = 'RA---TAN'           / R.A. in tangent plane projection
#CRPIX1  =     1044.60927497064 / Ref pix of axis 1
#CRVAL1  =     308.611677174582 / RA at Ref pix in decimal degrees
#CTYPE2  = 'DEC--TAN'           / DEC. in tangent plane projection
#CRPIX2  =     2289.53839638225 / Ref pix of axis 2
#CRVAL2  =     28.2797249736673 / DEC at Ref pix in decimal degrees
#CD1_1   = -2.0204567874974E-05 / WCS matrix element 1 1
#CD1_2   =  1.8157912081056E-08 / WCS matrix element 1 2
#CD2_1   = 2.39303378045856E-08 / WCS matrix element 2 1
#CD2_2   = 2.02198294404948E-05 / WCS matrix element 2 2

  sciext="["//l_sci_ext//"2]"
  imhead(l_inim[n_i]//sciext,long+) | match("CTYPE","STDIN",stop-, > filelist)
  imhead(l_inim[n_i]//sciext,long+) | match("CRPIX","STDIN",stop-, >> filelist)
  imhead(l_inim[n_i]//sciext,long+) | match("CRVAL","STDIN",stop-, >> filelist)
  imhead(l_inim[n_i]//sciext,long+) | match("CD","STDIN",stop-, >> filelist)
  imhead(l_inim[n_i]//sciext,long+) | match("GAIN","STDIN",stop-, >> filelist)
  imhead(l_inim[n_i]//sciext,long+) | match("RDNOISE","STDIN",stop-, >> filelist)

  mkheader(l_outim[n_i]//"[1]",filelist,append+,verbose-)  
  mkheader(l_outim[n_i]//"[0]",filelist,append+,verbose-)
  delete(filelist,verify-, >>& "dev$null")
  hedit(l_outim[n_i]//"[0]","CCDNAME",delete+,verify-,show-)
  hedit(l_outim[n_i]//"[1]","CCDNAME",delete+,verify-,show-)
  hedit(l_outim[n_i]//"[0]","CCDSIZE",delete+,verify-,show-)
  hedit(l_outim[n_i]//"[1]","CCDSIZE",delete+,verify-,show-)
  hedit(l_outim[n_i]//"[0]","CCDSEC",delete+,verify-,show-)
  hedit(l_outim[n_i]//"[1]","CCDSEC",delete+,verify-,show-)
  # Apply shift  - 32 pix overscan independent of binning
  imgets(l_outim[n_i]//"[1]","CRPIX1")
  hedit(l_outim[n_i]//"[1]","CRPIX1",str(real(imgets.value)+2086./Xbin-32.),add+,upd+,ver-,show-)
  hedit(l_outim[n_i]//"[0]","CRPIX1",str(real(imgets.value)+2086./Xbin-32.),add+, upd+,ver-,show-)
  gemhedit(l_outim[n_i]//"[0]","GMOSAIC",l_struct,"Time stamp for GMOSAIC")
  gemhedit(l_outim[n_i]//"[0]","GEM-TLM",l_struct,"Last modification with GEMINI")

  if(!l_fl_paste) {
     imdelete(tmpim[1]//","//tmpim[3],ver-, >>& "dev$null")
  }

}  # end of loop through images

goto clean

crash:
    status=1
    goto clean

clean:
    if(!l_fl_paste) {
       imdelete(tmpim[1]//","//tmpim[3],ver-, >>& "dev$null")
    }
    scanfile=""
    delete(filelist//","//mdf,ver-, >>& "dev$null")

printlog("----------------------------------------------------------------------------",l_logfile,l_verbose)

end
