# Copyright(c) 2002 Association of Universities for Research in Astronomy, Inc.

procedure gscut (inimage)

# Cut GMOS MOS spectra into 2D pieces
#
# Version Feb 28, 2002  BM  v1.3 release

string inimage   {"",prompt="Input image"}
string outimage  {"",prompt="Output image"}
string secfile   {"",prompt="Output file for image sections"}
bool   fl_vardq  {no,prompt="Propagate variance and data quality planes?"}
string gratingdb {"gmos$data/GMOSgratings.dat",prompt="Gratings database file"}
string filterdb  {"gmos$data/GMOSfilters.dat",prompt="Filters database file"}
string logfile   {"",prompt="Logfile"}
bool   verbose   {yes,prompt="Verbose output?"}
int    status    {0,prompt="Exit status (0=good)"}

begin

# local variables for task parameters
string l_inimage,l_outimage,l_title,l_logfile,l_secfile
string l_gratingdb,l_filterdb
real l_yoff
bool l_verbose,l_vardq

# other variables
string imgsec,sec,datsec,detsec,mdf,impaste
string sci,var,dq,modobs
string grating,filter,gname,fname,ffile,outroot,slittype
real cwave,grule,gblaze,gR,gcoverage,nmppx,xscale,yscale
real gwave1,gwave2,fwave1,fwave2,wave1,wave2,a,tilt,gtilt,greq,pi
real xccd,yccd,slitx,slity,slitwid,slitlen,pixcwave,refpix,asecmm
real ssize_mx,ssize_my,ssize_mr,ssize_mw,stilt_m,spos_mx,spos_my
real wavoffset
int xbin,ybin,speclen,specwid,center,xcen,ycen
int sposy,slitsepp,objid,nxpix,nypix
int nslit,j1,id1,nr,x1,y1,x2,y2,j,nextnd
bool delpaste
struct sdate

# Query parameters
l_inimage=inimage ; l_outimage=outimage #; l_title=title 
l_logfile=logfile
l_secfile=secfile
l_verbose=verbose
l_vardq=fl_vardq
l_gratingdb=gratingdb
l_filterdb=filterdb
status=0
delpaste=no

pi=3.14159265
asecmm=1.611444

# Keep imgets parameters from changing by outside world
cache("imgets", "hedit", "tinfo","fparse","tabpar","gmos","gimverify")

# Define temporary files
imgsec=mktemp("tmpimgsec")
sci=mktemp("tmpsci")
var=mktemp("tmpvar")
dq=mktemp("tmpdq")
mdf=mktemp("tmpmdf")
impaste=mktemp("tmppaste")

# Start logging to file
if (l_logfile == "STDOUT") {
   l_logfile = ""
   l_verbose=yes
}

# Test the logfile:
if (l_logfile == "" || stridx(" ",l_logfile)>0) {
    l_logfile = gmos.logfile
    if (l_logfile == "" || stridx(" ",l_logfile)>0) {
        l_logfile = "gmos.log"
        printlog("WARNING - GSCUT: Both gscut.logfile and gmos.logfile fields are empty",l_logfile,yes)
        printlog("                    Using default file gmos.log",l_logfile,yes)
    }
}

date | scan(sdate)
printlog("--------------------------------------------------------------------------------",l_logfile,verbose=l_verbose)
printlog("GSCUT -- "//sdate,l_logfile,verbose=l_verbose)
printlog(" ",l_logfile,verbose=l_verbose)
printlog("inimage = "//l_inimage,l_logfile,verbose=l_verbose)
printlog("outimage = "//l_outimage,l_logfile,verbose=l_verbose)
printlog("secfile = "//l_secfile,l_logfile,verbose=l_verbose)
printlog(" ",l_logfile,verbose=l_verbose)

#check that there are input files
if (l_inimage == "" || l_inimage == " "){
  printlog("ERROR - GSCUT: input files not specified",l_logfile,yes)
  goto error
}

# Check input image
gimverify(l_inimage)
if (gimverify.status!=0) {
    printlog("ERROR - GSCUT: input file doesn't exist or is not a MEF",l_logfile,yes)
    goto error
}
l_inimage=gimverify.outname//".fits"

# check existence of list file
if (substr(l_inimage,1,1) == "@") {
    printlog("ERROR - GSCUT: lists are currently not supported",l_logfile,yes)
    goto error
}

# check output name
if (l_outimage != "") {
    fparse(l_outimage,verbose-)
    outroot=fparse.root
    if (fparse.extension == "") {
        l_outimage=l_outimage//".fits"
    }

    # Check that the output file does not already exist. If so, exit.
    if (imaccess(l_outimage)) {  
        printlog("ERROR - GSCUT: Output file "//l_outimage//" already exists.",l_logfile,yes)
        goto error
    }
}

if (l_secfile != "" && access(l_secfile)) {  
    printlog("ERROR - GSCUT: Section file "//l_secfile//" already exists.",l_logfile,yes)
    goto error
}

# Check that the MDF exists
tinfo(l_inimage//"[MDF]",ttout-, >>& "dev$null")
if (tinfo.tbltype!="fits") {
    printlog("ERROR - GSCUT: MDF file does not exist",l_logfile,yes)
    goto error
}

# Paste the MEF together
imgets(l_inimage//"[0]","GMOSAIC", >>& "dev$null")
if (imgets.value == "0" && l_outimage!="") {
    print("Mosaicing")
    gmosaic(l_inimage,outimage=impaste,outpref="",gap=37,fl_paste=no,
    verb-,key_datsec="DATASEC",sci_ext="SCI",key_ccdsum="CCDSUM", 
    key_detsec="DETSEC")
    delpaste=yes
} else {
    impaste=l_inimage
    delpaste=no
}

# Find slits
# get header information
imgets(impaste//"[0]","GRATING")
grating=imgets.value

imgets(impaste//"[SCI,1]","CCDSUM")
print(imgets.value) | scan(xbin,ybin)
xscale=0.0727*real(xbin)
yscale=0.0727*real(ybin)

imgets(impaste//"[0]","FILTER2")
filter=imgets.value

imgets(impaste//"[0]","GRWLEN")
cwave=real(imgets.value)

imgets(impaste//"[0]","GRTILT")
tilt=real(imgets.value)*pi/180.

imgets(impaste//"[sci,1]","i_naxis1")
nxpix=int(imgets.value)

imgets(impaste//"[sci,1]","i_naxis2")
nypix=int(imgets.value)

# get grating information
match(grating,l_gratingdb,stop-,print+,meta+) | scan(gname,grule,gblaze,gR,gcoverage,gwave1,gwave2,wavoffset,l_yoff)

l_yoff=l_yoff/ybin
print(((cwave*grule)/1.e6)) | interp("gmos$data/gratingeq.dat","STDIN",int_mode="spline",curve_gen-) | scan(greq,gtilt)

#greq=(cwave*grule)/1.e6
gtilt=gtilt*pi/180.
a=sin(gtilt+0.872665)/sin(gtilt)
gR=206265.*greq/(0.5*81.0*sin(gtilt))
nmppx=a*0.0727*real(xbin)*cwave*81.0*sin(gtilt)/(206265.*greq)
#gcoverage=nmppx*real(nxpix)
#wave1=cwave-gcoverage/2.
#wave2=wave1+gcoverage

wave1=gwave1
wave2=gwave2

printlog("Central wavelength "//cwave,l_logfile,l_verbose)
printlog("Grating: "//gname,l_logfile,l_verbose)
printlog("Resolution (0.5'' slit): "//gR,l_logfile,l_verbose)
printlog("Anamorphic factor: "//a,l_logfile,l_verbose)
printlog("Grating tilt (header): "//(tilt*180./pi),l_logfile,l_verbose)
printlog("Calculated tilt: "//(gtilt*180./pi),l_logfile,l_verbose)
printlog("nm/pix = "//nmppx,l_logfile,l_verbose)

# get filter information
fwave1=0.0
fwave2=99999.0
if (filter != "" && filter != "open2-6") {
    match(filter,l_filterdb,stop-,print+,meta+) | scan(fname,fwave1,fwave2,ffile)
}

# determine whether filter or grating limits wavelength coverage
if (fwave1 > wave1) {
    wave1 = fwave1
}
if (fwave2 < wave2) {
    wave2 = fwave2
}
# and red limit of CCD
if (wave2 > 1050.) {
    wave2 = 1050.
}


# in pixels
speclen=nint((wave2-wave1)/nmppx)
printlog("Filter: "//filter,l_logfile,l_verbose)
printlog("Wavelength coverage: "//wave1//" "//wave2,l_logfile,l_verbose)
printlog("Spectrum length = "//speclen,l_logfile,l_verbose)

# pixel value of central wavelength from left (red) end of spectrum
pixcwave=speclen-(cwave-wave1)/nmppx
printlog("Pixel of central wavelength: "//pixcwave,l_logfile,l_verbose)

# Copy MDF to output image
if (l_outimage != "") {
    # slice out the MDF
    tcopy(l_inimage//"[MDF]",mdf//".fits",verbose-)
    wmef(mdf//".fits",l_outimage,extnames="MDF",verb-,phu=l_inimage, >>& "dev$null")
} 

# Add section and select columns to MDF
tcalc(l_outimage//"[MDF]","SELECT",1,datatype="int")
tcalc(l_outimage//"[MDF]","SECX1",1,datatype="int")
tcalc(l_outimage//"[MDF]","SECX2",1,datatype="int")
tcalc(l_outimage//"[MDF]","SECY1",1,datatype="int")
tcalc(l_outimage//"[MDF]","SECY2",1,datatype="int")

# How many slits to extract? If LONGSLIT nr=1.
imgets(l_inimage//"[0]","OBSMODE")
modobs=imgets.value

if(modobs == "LONGSLIT") {
  nr=1
} else if(modobs == "MOS") {
  tinfo(l_inimage//"[MDF]",ttout-)
  nr=tinfo.nrows
}

# Loop over slits
slittype="" ; slitlen=0.0
xcen=real(nxpix)/2.
ycen=real(nypix)/2.
for (i=1; i<=nr; i+=1) {
  if (modobs == "MOS") {
    # Get parameters for slit from MDF (for MOS only)
    tprint(l_inimage//"[MDF]",prparam-,prdata+,showrow-,showhdr-,showunits-,\
    col="slitid,slitpos_mx,slitpos_my,slitsize_mx,slitsize_my,slitsize_mr,\
    slitsize_mr,slitsize_mw,slittilt_m,SLITTYPE",\
    rows=i,pwidth=160) | scan(objid,spos_mx,spos_my,ssize_mx,ssize_my,\
    ssize_mr,ssize_mw,stilt_m,slittype)
    
    #Convert from mask to pixel coordinates and correct for binning in both
    # directions
    xccd=spos_mx*asecmm/xscale
#    yccd=spos_my*asecmm/yscale
    yccd=0.99591859227*spos_my + 5.3042211333437E-8*spos_my**2 + 1.7447902551997E-7*spos_my**3
    yccd=yccd*asecmm/yscale
    slitwid=ssize_mx*asecmm
    slitlen=ssize_my*asecmm
    
    # set slit length if the aperture is a circle, radius=slitwid
    if (slittype=="circle") {
        slitlen=2.*slitwid
    }

    xccd=xcen+xccd
    yccd=ycen+yccd

    # slit height
    specwid=nint(1.05*slitlen/yscale)
    printlog("Slit "//i//": slit height = "//specwid,l_logfile,l_verbose)
    center=specwid/2

    # offset in y
    #sposy=nint(slity/scale)
  } else if (modobs == "LONGSLIT") {
    specwid=real(nypix)
    center=specwid/2.
    xccd=xcen
    yccd=center
    l_yoff=0.0
  }

  refpix=pixcwave

  # Position of object, take into account that lambda decreases with x
  x1=nint(xcen-(xcen-xccd)/a-pixcwave)+wavoffset/nmppx
  #+0.000189*(yccd-2304./ybin) - 1.044231E-5*(yccd-2304./ybin)**2
  x2=x1+speclen-1
  y1=nint(yccd-center+l_yoff)
  y2=y1+specwid
  # check spectrum isn't off chip
  if (x1 < 1) {
      refpix=refpix+real(x1-1)
      x1=1
  }
  if (x2 > nxpix) {
      x2=nxpix
  }
  if (y1 < 1) {
      y1=1
  }
  if (y2 > nypix) {
      y2=nypix
  }
#  } else if (modobs == "LONGSLIT") {
#    x1=1
#    x2=nxpix
#    y1=1
#    y2=nypix
#    pixcwave=pixcwave+wavoffset/nmppx
#  }
    
  sec="["//x1//":"//x2//","//y1//":"//y2//"]"
  datsec="[1:"//(x2-x1+1)//",1:"//(y2-y1+1)//"]"
  detsec="["//x1//":"//x2//","//y1//":"//y2//"]"
  if (l_secfile != "") {
      print(sec, >> l_secfile)
  }

  if (l_outimage != "") {
      imgsec=outroot//"_"//i
      if (imaccess(imgsec)) {
          imdelete(imgsec,verify-)
      }

      # Put image section info in MDF columns
      if (modobs == "MOS") {
        partab(x1,l_outimage//"[MDF]","SECX1",i)
        partab(x2,l_outimage//"[MDF]","SECX2",i)
        partab(y1,l_outimage//"[MDF]","SECY1",i)
        partab(y2,l_outimage//"[MDF]","SECY2",i)
      }
      # Cut this slit and insert in output file
      imcopy(impaste//"[SCI,1]"//sec,sci,verbose-)
      imgets(l_outimage//"[0]","NEXTEND")
      nextnd=int(imgets.value)
      fxinsert(sci//".fits",l_outimage//"["//nextnd//"]","0",verbose-, >>& "dev$null")
      nextnd=nextnd+1
      hedit(l_outimage//"[0]","NEXTEND",nextnd,update+,verify-,show-)
      hedit(l_outimage//"["//nextnd//"]","EXTNAME","SCI",add+,update+,verify-,show-)
      hedit(l_outimage//"["//nextnd//"]","EXTVER",i,add+,update+,verify-,show-)
      gemhedit(l_outimage//"["//nextnd//"]","MDFROW",i,"Corresponding row in MDF")
      hedit(l_outimage//"["//nextnd//"]","DETSEC",detsec,add+,update+,verify-,show-)
      hedit(l_outimage//"["//nextnd//"]","DATASEC",datsec,add+,update+,verify-,show-)
      hedit(l_outimage//"["//nextnd//"]","CCDSUM",xbin//" "//ybin,add+,update+,verify-,show-)

      #update header with wavelength information
      gemhedit(l_outimage//"["//nextnd//"]","REFPIX1",refpix,"Pixel of central wavelength")
      # Variance and DQ planes
      if (l_vardq) {
          if (imaccess(impaste//"[VAR,1]")) {
              imcopy(impaste//"[VAR,1]"//sec,var,verbose-)
              imcopy(impaste//"[DQ,1]"//sec,dq,verbose-)
          } else {
              imarith(sci,"*",sci,var,verb-)
              imarith(sci,"*",0,dq,pixtype="int",calctype="int",verb-)
          }
          # Insert
          fxinsert(var//".fits,"//dq//".fits",l_outimage//"["//nextnd//"]","0",verbose-, >>& "dev$null")
          hedit(l_outimage//"[0]","NEXTEND",(nextnd+2),update+,verify-,show-)
          # Variance header
          hedit(l_outimage//"["//(nextnd+1)//"]","EXTNAME","VAR",add+,update+,verify-,show-)
          hedit(l_outimage//"["//(nextnd+1)//"]","EXTVER",i,add+,update+,verify-,show-)
          gemhedit(l_outimage//"["//(nextnd+1)//"]","MDFROW",i,"Corresponding row in MDF")
          gemhedit(l_outimage//"["//(nextnd+1)//"]","REFPIX1",refpix,"Referenc pixel of central wavelength")
          hedit(l_outimage//"["//(nextnd+1)//"]","DETSEC",detsec,add+,update+,verify-,show-)
          hedit(l_outimage//"["//(nextnd+1)//"]","DATASEC",datsec,add+,update+,verify-,show-)
          hedit(l_outimage//"["//(nextnd+1)//"]","CCDSUM",xbin//" "//ybin,add+,update+,verify-,show-)
          # DQ header
          hedit(l_outimage//"["//(nextnd+2)//"]","EXTNAME","DQ",add+,update+,verify-,show-)
          hedit(l_outimage//"["//(nextnd+2)//"]","EXTVER",i,add+,update+,verify-, show-)
          gemhedit(l_outimage//"["//(nextnd+2)//"]","MDFROW",i,"Corresponding row in MDF")
          hedit(l_outimage//"["//(nextnd+2)//"]","REFPIX1",refpix,"Reference pixel of central wavelength")
          hedit(l_outimage//"["//(nextnd+2)//"]","DETSEC",detsec,add+,update+,verify-,show-)
          hedit(l_outimage//"["//(nextnd+2)//"]","DATASEC",datsec,add+,update+,verify-,show-)
          hedit(l_outimage//"["//(nextnd+2)//"]","CCDSUM",xbin//" "//ybin,add+,update+,verify-,show-)
      }
      imdelete(imgsec//","//sci//","//var//","//dq,verify-, >>& "dev$null")
  }
}

# final header update
if(l_outimage!="") {
  hedit(l_outimage//"[0]","NSCIEXT",nr,add+,update+,verify-,show-)
  gemhedit(l_outimage//"[0]","GSCUT",sdate,"Time stamp for GSCUT")
  gemhedit(l_outimage//"[0]","GEM-TLM",sdate,"Last edit with GEMINI")
}

# clean up
goto clean

error:
    status=1
    goto clean

clean:
    delete("tmpslit*.fits,"//mdf//".fits",verify-, >>& "dev$null")
    imdelete(imgsec//","//sci//","//var//","//dq,verify-, >>& "dev$null")
    if (delpaste) {
        imdelete(impaste,verify-)
    }
    # close log file
    printlog(" ",l_logfile,l_verbose)
    printlog("GSCUT done",l_logfile,l_verbose)
    printlog("--------------------------------------------------------------------------------",l_logfile,l_verbose )

end
