# Copyright(c) 2002 Association of Universities for Research in Astronomy, Inc.

procedure gstransform(inimages)

# Rectify GMOS spectra using available S-distortion correction. 
#
# Version   Feb 28, 2002  ML,BM  v1.3 release

string inimages     {prompt="Input GMOS spectra"}
string outimages    {"",prompt="Output spectra"}
string outprefix    {"t",prompt="Prefix for output spectra"}
bool   fl_stran     {no,prompt="Apply S-distortion correction"}
string sdistname    {"",prompt="Names of S-distortions calibrations"}
bool   fl_wavtran   {yes,prompt="Apply wavelength calibration from arc spectrum"}
string wavtraname   {"",prompt="Names of wavelength calibrations"} 
string database     {"database",prompt="Directory for calibration files"}
bool   fl_vardq     {no,prompt="Transform variance and data quality planes"} 
string interptype   {"linear",enum="nearest|linear|poly3|poly5|spline3",prompt="Interpolation type for transform"} 
bool   xlog         {no,prompt="Logarithmic x coordinate for transform"}
bool   ylog         {no,prompt="Logarithmic y coordinate for transform"}
bool   fl_flux      {yes,prompt="Conserve flux per pixel in the transform"} 
string  sci_ext     {"SCI",prompt="Name of science extension"}
string  var_ext     {"VAR",prompt="Name of variance extension"}
string  dq_ext      {"DQ",prompt="Name of data quality extension"}
string  logfile     {"",prompt="Logfile"}
#bool    fl_inter    {yes,prompt="Interactive?"}
bool    verbose     {yes,prompt="Verbose"}
int     status      {0,prompt="Exit status (0=good)"}
struct* scanfile    {"",prompt="Internal use"}

begin

#Local variable declarations 
string l_inimages, l_outimages, l_prefix, l_sdistname, l_wavtran, l_wavtraname
string l_grating, l_filter, l_gratingdb, l_filterdb
string l_database, l_interptype, l_sci_ext, l_var_ext, l_dq_ext
string l_logfile, l_fitcfunc
bool l_fl_vardq, l_fl_wavtran, l_fl_stran
bool l_fl_inter, l_xlog, l_ylog, l_flux, l_verbose 
int  l_spatbin, l_specbin, l_fitcxord, l_fitcyord, l_status 


#Other variables used within this task 

file temp1,temp2,temp3,wavlist,sdistlist,junk,tmpsci,tmpvar,tmpdq,mdf
string inlist,outlist,img,outimg,snum,infile[100]
string wcalfile,scalfile,wcal,scal,fitname
string obsmode,dbprefix, suf, l_stdout
string ffile,fname,gname
int nbad,nin,nout,nwavfil,ndistfil,nsciext,nextens,i  
int nsciin,nsciout,sel,nextnd,mdfrow
bool useprefix,go
struct sdate,wlist  

# Set the local variables
l_inimages=inimages ; l_outimages=outimages ; l_prefix=outprefix
l_fl_stran=fl_stran; l_sdistname=sdistname 
l_fl_wavtran=fl_wavtran ; l_wavtraname=wavtraname ; l_database=database
#l_fl_inter=fl_inter
l_fl_vardq=fl_vardq ; l_interptype=interptype ; l_xlog=xlog ; l_ylog=ylog 
l_flux=fl_flux
l_sci_ext=sci_ext ; l_var_ext=var_ext ; l_dq_ext=dq_ext 
l_logfile=logfile ; l_verbose=verbose 

# Initialize exit status
status=0

l_stdout=""
if (l_verbose) {
   l_stdout="STDOUT,"
}

# cache some parameter files
cache("imgets","gmos","hedit","gimverify")

# Check logfile
if (l_logfile == "" || stridx(" ",l_logfile)>0) {
    l_logfile = gmos.logfile
    if (l_logfile == "" || stridx(" ",l_logfile)>0) {
        l_logfile = "gmos.log"
        printlog("WARNING - GSTRANSFORM: Both gstransform.logfile and gmos.logfile fields are empty",l_logfile,yes)
        printlog("                    Using default file gmos.log",l_logfile,yes)
    }
}

# temporary files
wavlist=mktemp("tmpwavlist") 
sdistlist=mktemp("tmpsdistlist")
temp3=mktemp("tmpfilelist")
junk=mktemp("junk") 

printlog("-------------------------------------------------------------------------------",l_logfile,l_verbose)
date | scan(sdate)
printlog("GSTRANSFORM -- "//sdate,l_logfile,l_verbose)
print(" ", >> l_logfile)


#The usual kind of checks to make sure everything we need is specified. 
nbad=0

if(l_inimages=="" || l_inimages==" ") {
  printlog("ERROR - GSTRANSFORM: Input spectra is an empty string",l_logfile,yes)
  nbad=nbad+1 
}

if(substr(l_inimages,1,1)=="@") {
   inlist=substr(l_inimages,2,strlen(l_inimages))
   if(!access(inlist)) {
      printlog("ERROR - GSTRANSFORM: Input list "//inlist//" not found",l_logfile,verbose+)
      nbad=nbad+1
   }
}

if((l_outimages=="" || l_outimages==" ") && (l_prefix=="" || l_prefix==" ")) {
    printlog("ERROR - GSTRANSFORM: Neither the output spectra nor prefix is specified.",l_logfile,yes)
    nbad=nbad+1
} else if ((l_outimages!="" && l_outimages!=" ")) {
    useprefix=no
} else {
    useprefix=yes
}

if(substr(l_outimages,1,1)=="@") {
   outlist=substr(l_outimages,2,strlen(l_outimages))
   if(!access(outlist)) {
      printlog("ERROR - GSTRANSFORM: Output list "//outlist//" not found",l_logfile,yes)
      nbad=nbad+1
   }
}

temp1=mktemp("tmpin")
temp2=mktemp("tmpout") 
files(l_inimages,sort-, > temp1)
count(temp1) | scan(nin)
if(!useprefix) { 
    files(l_outimages,sort-, > temp2)
    count(temp2) | scan(nout)
    if (nin != nout) {
        printlog("ERROR - GSTRANSFORM: Different number of input and output spectra",l_logfile,yes)
        nbad=nbad+1
    }
} else {
   files(l_prefix//"//@"//temp1,sort-, > temp2)
}

# Are we going to do anything?
if(l_fl_wavtran==no && l_fl_stran==no) { 
    printlog("ERROR - GSTRANSFORM: Neither S-distortion or wavelength correction have been requested",l_logfile,yes)
    nbad=nbad+1
} 


# S-distortion calibrations
if(l_fl_stran) { 
    if(substr(l_sdistname,1,1)=="@") {
        soutlist=substr(l_sdistname,2,strlen(l_sdistname))
        if(!access(soutlist)) {
            printlog("ERROR - GSTRANSFORM: Output list "//soutlist//" not found",l_logfile,yes)
            nbad=nbad+1
        }
    }
    if (nbad==0) {
        files(l_sdistname,sort-, > sdistlist)
        count(sdistlist) | scan(ndistfil)
        if(ndistfil != 1 || ndistfil != nin) { 
            printlog("ERROR - GSTRANSFORM: Number of S-distortion images must be 1, or ",l_logfile,yes)
            printlog("                     equal to the number of input images.",l_logfile,yes)
            nbad=nbad+1
        } 
        if (ndistfil == 1 && nbad==0) {
            delete(sdistlist,verify-)
            for(i=1;i<=nin;i+=1) { 
                print(l_sdistname, >> sdistlist) 
            } 
        }
    }
} else {
    for(i=1;i<=nin;i+=1) { 
        print("none", >> sdistlist) 
    } 
}

#Wavelength calibrations
if(l_fl_wavtran) { 
    if(substr(l_wavtraname,1,1)=="@") {
        woutlist=substr(l_wavtraname,2,strlen(l_wavtraname))
        if(!access(woutlist)) {
            printlog("ERROR - GSTRANSFORM: Output list "//woutlist//" not found",l_logfile,yes)
            nbad=nbad+1
        }
    }
    if (nbad == 0) { 
        files(l_wavtraname,sort-, > wavlist)
        count(wavlist) | scan(nwavfil)
        if(nwavfil != 1 && nwavfil != nin) { 
            printlog("ERROR - GSTRANSFORM: Number of wavelength calibration images must be 1, or ",l_logfile,yes)
            printlog("                     equal to the number of input images.",l_logfile,yes)
            nbad=nbad+1
        } 
        if (nwavfil == 1 && nbad==0) {
            delete(wavlist,verify-)
            for(i=1;i<=nin;i+=1) { 
                print(l_wavtraname, >> wavlist) 
            } 
        }
    }
} else {
    for(i=1;i<=nin;i+=1) { 
        print("none", >> wavlist) 
    } 
}
   
if(l_sci_ext=="" || l_sci_ext==" ") {
  printlog("ERROR - GSTRANSFORM: extension name sci_ext is missing",l_logfile,yes)
  nbad=nbad+1
}

#If var propogation is requested, make sure the names are given
if(l_fl_vardq) {
   if (l_dq_ext=="" || l_dq_ext ==" ") {
        printlog("ERROR - GSTRANSFORM: extension name dq_ext is missing",l_logfile,yes)
        nbad=nbad+1
   } else if (l_var_ext=="" || l_var_ext ==" ") {
        printlog("ERROR - GSTRANSFORM: extension name var_ext is missing",l_logfile,yes)
        nbad=nbad+1
   }
}

# check no commas in sci_ext, var_ext and dq_ext
if(stridx(",",l_sci_ext)>0 || stridx(",",l_var_ext)>0 || stridx(",",l_dq_ext)>0 ) {
  printlog("ERROR - GSTRANSFORM: sci_ext, var_ext or dq_ext contains commas, give root name only",l_logfile,yes)
  nbad=nbad+1
}

i=0
scanfile=temp1
while(fscan(scanfile,img) !=EOF) { 
    gimverify(img)
    if (gimverify.status>0) {
         nbad=nbad+1
    } else {
        i=i+1
        # name w/o suffix
        infile[i]=gimverify.outname
    }
#end of while loop over images. 
}
scanfile=""

#If anything was wrong then exit. 
if (nbad > 0) { 
    printlog("ERROR - GSTRANSFORM: "//nbad//" errors found with input parameters. Exiting.",l_logfile,yes) 
    goto error
} 

   
#If we are here then everything should be OK. 
#Write all the relevant info to the logfile:
#
printlog("",l_logfile,l_verbose)
printlog("inimages   = "//l_inimages,l_logfile,l_verbose) 
printlog("outimages  = "//l_outimages,l_logfile,l_verbose) 
printlog("outprefix  = "//l_prefix,l_logfile,l_verbose) 
printlog("fl_stran   = "//l_fl_stran,l_logfile,l_verbose)
if(l_fl_stran)
   printlog("sdistname  = "//l_sdistname,l_logfile,l_verbose)
printlog("fl_wavtran = "//l_fl_wavtran,l_logfile,l_verbose)
if(l_fl_wavtran)
printlog("wavtraname = "//l_wavtraname,l_logfile,l_verbose)
printlog("database   = "//l_database,l_logfile,l_verbose)
#printlog("fl_inter = "//l_fl_inter,l_logfile,l_verbose)
printlog("fl_vardq   = "//l_fl_vardq,l_logfile,l_verbose)
printlog("interptype = "//l_interptype,l_logfile,l_verbose)
printlog("xlog       = "//l_xlog,l_logfile,l_verbose)
printlog("ylog       = "//l_ylog,l_logfile,l_verbose)
printlog("fl_flux    = "//l_flux,l_logfile,l_verbose)
printlog("sci_ext    = "//l_sci_ext,l_logfile,l_verbose)
if(l_fl_vardq) { 
   printlog("var_ext = "//l_var_ext,l_logfile,l_verbose)
   printlog("dq_ext  = "//l_dq_ext,l_logfile,l_verbose)
} 
printlog("",l_logfile,l_verbose)

#
#Join lists together. The logic here is a tad confusing but it generates 
#a structured filelist which includes the input file, outputfile name (MEF), 
#and the transformation database(s) to use for each image in the input list - 
#all on one line per image to be transformed. 
#
  
# Make list file for main loop
join(temp1//","//temp2//","//sdistlist//","//wavlist,delim=" ",missing="Missing",maxchar=161,shortest=yes,verbose=no, > temp3)
scanfile=temp3 
while(fscan(scanfile,img,outimg,scal,wcal) !=EOF) { 
    # check .fits
    suf=substr(img,strlen(img)-4,strlen(img))
    if (suf!=".fits") {
      img=img//".fits"
    }

    suf=substr(outimg,strlen(outimg)-4,strlen(outimg))
    if (suf!=".fits") {
      outimg=outimg//".fits"
    }

    # copy MDF
    mdf=mktemp("tmpmdf")
    tcopy(img//"[MDF]",mdf//".fits",verbose-)
    wmef(mdf//".fits",outimg,extnames="MDF",verb-,phu=img, >& "dev$null")
    imdelete(mdf,verify-)
    
    # get obsmode
    imgets(img//"[0]","OBSMODE")
    obsmode=imgets.value
    
    # get number of rows in MDF
    imgets(img//"[0]","NSCIEXT")
    nsciin=int(imgets.value)
    # Loop over extensions
    nsciout=0
    for (i=1; i<=nsciin; i+=1) {
      sel=1
      mdfrow=i
      if (obsmode == "MOS") {
          imgets(img//"["//l_sci_ext//","//i//"]","MDFROW", >& "dev$null")
          if (imgets.value != "0") {
              mdfrow=int(imgets.value)
              tabpar(img//"[MDF]","SELECT",mdfrow)
              if (tabpar.undef==no) {
                  sel=int(tabpar.value)
              }
          }
      } 
      if (sel==1) {
        if (mdfrow < 10) {
            snum="00"//mdfrow
        } else if (mdfrow < 100) {
            snum="0"//mdfrow
        } else {
            snum=str(mdfrow)
        }
        if (obsmode=="IFU") {
            dbprefix="id"
            l_fl_stran=no
        } else {
            dbprefix="fc"
        }
        scalfile=scal//"_"//snum
        wcalfile=wcal//"_"//snum
        if (l_fl_stran && l_fl_wavtran) {
            fitname=scalfile//","//wcalfile
            go=access(l_database//"/"//dbprefix//wcalfile)
            go=access(l_database//"/"//dbprefix//scalfile)
        } else if (l_fl_stran && !l_fl_wavtran) {
            fitname=scalfile
            go=access(l_database//"/"//dbprefix//scalfile)
        } else if (!l_fl_stran && l_fl_wavtran) {
            fitname=wcalfile
            go=access(l_database//"/"//dbprefix//wcalfile)
        }
        if (!go) {
            printlog("WARNING - GSTRANFORM: wavelength or S-distortion calibration file not found",l_logfile,yes)
        }
        if (go && imaccess(img//"["//l_sci_ext//","//i//"]")) {
            nsciout=nsciout+1
            printlog("Transforming "//img//"["//l_sci_ext//","//i//"]",l_logfile,l_verbose)
            tmpsci=mktemp("tmpsci") 
            tmpvar=mktemp("tmpvar") 
            tmpdq=mktemp("tmpdq") 

            if (obsmode == "IFU") {
                # science plane
                hedit(img//"["//l_sci_ext//","//i//"]","REFSPEC1",
                wcalfile,update+,add+,verify-)
                dispcor(img//"["//l_sci_ext//","//i//"]",tmpsci,linear+,\
                database=l_database,table="",w1=INDEF,w2=INDEF,dw=INDEF,\
                nw=INDEF,log-,flux=l_flux,samedisp+,global-,ignoreaps-,\
                listonly-,verbose=l_verbose,logfile=l_logfile)
                if (l_fl_vardq && imaccess(img//"["//l_var_ext//","//i//"]")) {
                    # variance plane
                    hedit(img//"["//l_var_ext//","//i//"]","REFSPEC1",
                    wcalfile,update+,add+,verify-)
                    dispcor(img//"["//l_var_ext//","//i//"]",tmpvar,linear+,
                    database=l_database,table="",w1=INDEF,w2=INDEF,dw=INDEF,
                    nw=INDEF,log-,flux=l_flux,samedisp+,global-,ignoreaps-,
                    listonly-,verbose=l_verbose,logfile=l_logfile)
                    # dq plane
                    hedit(img//"["//l_dq_ext//","//i//"]","REFSPEC1",
                    wcalfile,update+,add+,verify-)
                    dispcor(img//"["//l_dq_ext//","//i//"]",tmpdq,linear+,
                    database=l_database,table="",w1=INDEF,w2=INDEF,dw=INDEF,
                    nw=INDEF,log-,flux=l_flux,samedisp+,global-,ignoreaps-,
                    listonly-,verbose=l_verbose,logfile=l_logfile)
                }
            } else {
                # science extension
                transform(img//"["//l_sci_ext//","//i//"]",tmpsci,
                fitnames=fitname,database=l_database,interptype=l_interptype,
                x1=INDEF,x2=INDEF,dx=INDEF,nx=INDEF,xlog=l_xlog,y1=INDEF,
                y2=INDEF,dy=INDEF,ny=INDEF,ylog=l_ylog,flux=l_flux,
                logfiles=l_stdout//l_logfile)
                if(l_fl_vardq && imaccess(img//"["//l_var_ext//","//i//"]")) { 
                    # variance plane
                    transform(img//"["//l_var_ext//","//i//"]",tmpvar,
                    fitnames=fitname,
                    database=l_database,interptype=l_interptype,x1=INDEF,
                    x2=INDEF,dx=INDEF,nx=INDEF,xlog=l_xlog, y1=INDEF,y2=INDEF,
                    dy=INDEF,ny=INDEF,ylog=l_ylog,flux=l_lflux,
                    logfiles=l_stdout//l_logfile) 
                    # dq plane
                    transform(img//"["//l_dq_ext//","//i//"]",tmpdq,
                    fitnames=fitname,
                    database=l_database,interptype=l_interptype,x1=INDEF,
                    x2=INDEF,dx=INDEF,nx=INDEF,xlog=l_xlog, y1=INDEF,y2=INDEF,
                    dy=INDEF,ny=INDEF,ylog=l_ylog,flux=l_lflux,
                    logfiles=l_stdout//l_logfile) 
                } 
            }
            ## Put transformed slit in output file 
            imgets(outimg//"[0]","NEXTEND")
            nextnd=int(imgets.value)
            fxinsert(tmpsci//".fits",outimg//"["//nextnd//"]","0",verbose-, >& "dev$null")
            nextnd=nextnd+1
            hedit(outimg//"[0]","NEXTEND",nextnd,update+,verify-,show-)
            hedit(outimg//"["//nextnd//"]","EXTNAME",l_sci_ext,add+,update+,verify-,show-)
            hedit(outimg//"["//nextnd//"]","EXTVER",nsciout,add+,update+,verify-,show-)
            if (l_fl_stran) {
                hedit(outimg//"["//nextnd//"]","SDISTNAM",scalfile,"Name of S-distortion")
            }
            if (l_fl_wavtran) {
                gemhedit(outimg//"["//nextnd//"]","WAVTRAN",wcalfile,"Name of wavelength transformation")
            }
            # Variance and DQ
            if (imaccess(tmpvar)) {
                fxinsert(tmpvar//".fits,"//tmpdq//".fits",outimg//"["//nextnd//"]","0",verbose-, >& "dev$null")
                hedit(outimg//"[0]","NEXTEND",(nextnd+2),update+,verify-,show-)
                # Variance header
                hedit(outimg//"["//(nextnd+1)//"]","EXTVER",nsciout,add+,update+,verify-,show-)
                if (l_fl_stran) {
                    hedit(outimg//"["//(nextnd+1)//"]","SDISTNAM",scalfile,"Name of S-distortion")
                }
                if (l_fl_wavtran) {
                    hedit(outimg//"["//(nextnd+1)//"]","WAVTRAN",wcalfile,"Name of wavelength transformation")
                }
                # DQ header
                hedit(outimg//"["//(nextnd+2)//"]","EXTVER",nsciout,add+,update+,verify-,show-)
                if (l_fl_stran) {
                    hedit(outimg//"["//(nextnd+2)//"]","SDISTNAM",scalfile,"Name of S-distortion")
                }
                if (l_fl_wavtran) {
                    hedit(outimg//"["//(nextnd+2)//"]","WAVTRAN",wcalfile,"Name of wavelength transformation")
                }
            }
            imdelete(tmpsci//","//tmpvar//","//tmpdq,verify-, >& "dev$null")         
         }
         hedit(outimg//"[0]","NSCIEXT",nsciout,update+,verify-,show-)
         gemhedit(outimg//"[0]","GEM-TLM",sdate,"Last modification with GEMINI")
         gemhedit(outimg//"[0]","GSTRANSF",sdate,"Time stamp for GSTRANSFORM")
       }
    } # end loop over extensions
} #end of while loop over input image list
scanfile=""
       
goto clean

error:
    status=1
    goto clean

#Clean Up 
clean: 
    printlog("",l_logfile,l_verbose)
    scanfile=""
    printlog("GSTRANSFORM done",l_logfile,l_verbose)
    printlog("-------------------------------------------------------------------------------",l_logfile,l_verbose)
    printlog("",l_logfile,l_verbose)
    delete(temp1//","//temp2//","//temp3//","//wavlist//","//sdistlist,verify-, >& "dev$null") 

end
